package main

import (
	"fmt"
	"strconv"

	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"encoding/json"
	"errors"
)

// SimpleChaincode example simple Chaincode implementation
type SimpleChaincode struct {
}

func (t *SimpleChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	fmt.Println("ex02 Init")

	return shim.Success(nil)
}

func (t *SimpleChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	fmt.Println("ex02 Invoke")
	function, args := stub.GetFunctionAndParameters()
	if function == "invoke" {
		// Make payment of X units from A to B
		return t.invoke(stub, args)
	} else if function == "add" {
		// Deletes an entity from its state
		return t.add(stub, args)
	} else if function == "delete" {
		// Deletes an entity from its state
		return t.delete(stub, args)
	} else if function == "query" {
		// the old "Query" is now implemtned in invoke
		return t.query(stub, args)
	}

	return shim.Error("Invalid invoke function name. Expecting \"invoke\" \"delete\" \"query\"")
}

// Transaction makes payment of X units from A to B
func (t *SimpleChaincode) invoke(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	var from, to string // Entities
	var X int           // Transaction value
	var err error

	if len(args) != 3 {
		return shim.Error("Incorrect number of arguments. Expecting 3")
	}

	// Perform the execution
	from = args[0]
	to = args[1]
	X, err = strconv.Atoi(args[2])
	if err != nil {
		return shim.Error("Invalid transaction amount, expecting a integer value")
	}

	// get old value
	fromVal, err := getBalance(stub, from)
	if err != nil {
		return shim.Error(err.Error())
	}
	toVal, err := getBalance(stub, to)
	if err != nil {
		return shim.Error(err.Error())
	}

	// put new value
	err = putBalance(stub, from, strconv.Itoa(fromVal - X))
	if err != nil {
		return shim.Error(err.Error())
	}
	err = putBalance(stub, to, strconv.Itoa(toVal + X))
	if err != nil {
		return shim.Error(err.Error())
	}

	return shim.Success(nil)
}

// Deletes an entity from state
func (t *SimpleChaincode) add(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 2 {
		return shim.Error("Incorrect number of arguments. Expecting 4")
	}
	user := args[0]
	value, err := strconv.Atoi(args[1])
	if err != nil {
		return shim.Error("Expecting integer value for asset holding")
	}
	err = putBalance(stub, user, strconv.Itoa(value))
	if err != nil {
		return shim.Error(err.Error())
	}

	return shim.Success(nil)
}

// Deletes an entity from state
func (t *SimpleChaincode) delete(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	if len(args) != 1 {
		return shim.Error("Incorrect number of arguments. Expecting 1")
	}

	user := args[0]
	sqlStr := "DELETE FROM account WHERE user = '" + user + "'"
	err := stub.DeleteStateBySql(sqlStr)
	if err != nil {
		return shim.Error(err.Error())
	}

	return shim.Success(nil)
}

// query callback representing the query of a chaincode
func (t *SimpleChaincode) query(stub shim.ChaincodeStubInterface, args []string) pb.Response {
	var user string // Entities
	var err error

	if len(args) != 1 {
		return shim.Error("Incorrect number of arguments. Expecting name of the person to query")
	}

	user = args[0]
	balance, err := getBalance(stub, user)
	if err != nil {
		return shim.Error(err.Error())
	} else {
		return shim.Success([]byte(strconv.Itoa(balance)))
	}
}

func getBalance(stub shim.ChaincodeStubInterface, user string) (int, error) {
	sqlStr := "SELECT balance FROM account WHERE user = '" + user + "'"
	sqlResult, err := stub.GetStateBySql(sqlStr)
	if err != nil {
		return 0, err
	}
	var jsonParsed [][]string
	json.Unmarshal(sqlResult, &jsonParsed)
	if len(jsonParsed) != 2 {
		return 0, errors.New("Do not contain user: " + user)
	} else {
		balance, err:= strconv.Atoi(jsonParsed[1][0])
		if err != nil {
			return 0, err
		} else {
			return balance, nil
		}
	}
}

func putBalance(stub shim.ChaincodeStubInterface, user string, value string) error {
	sqlStr := "INSERT INTO account (user, balance) VALUES ( " + user + ", " + value +")"
	err := stub.PutStateBySql(sqlStr)
	if err != nil {
		return err
	}
	return nil
}

func main() {
	err := shim.Start(new(SimpleChaincode))
	if err != nil {
		fmt.Printf("Error starting Simple chaincode: %s", err)
	}
}
